### TOPIC MODELING ###
## ## Hot Topics of Clinical Psychology ## ##


#### load data ----

# load preprocessed PSYNDEX Export
load("klinpsy_KP.RData") 
# data frame with abstracts (ABH), titles (TI), translated titles (TIU), controlled terms (CTd), 
# uncontrolled terms (UTG), additional terms (ITd), author information (YYKLAU), publication year (PY)



#### libraries ----
library(quanteda)
library(dplyr)
library(topicmodels)
library(quantqual) # devtools::install_github("AndreasFischer1985/quantqual")
library(parallel)
  

#### corpus creation ---- 

### a corpus of Titles, Abstracts, keywords and controlled terms will be used.
### all docs need to be in the same language.

### all docs have titles and controlled terms.
### some have abstacts and/or keywords.

table(!is.na(klinpsy_KP$ABH), !is.na(klinpsy_KP$ABN))
# 4758 docs (3.5 %) neither have abstract nor additional abstract


### new variable "text" is added that contains all existing text information 
### (abstracts, titles, (un)controlled terms, additional terms)
klinpsy_KP$text <- NA

# 1st case: abstracts and titles are in German
klinpsy_KP$text <- ifelse(klinpsy_KP$ABLH == "german", 
                        paste(klinpsy_KP$`TI+TIU`, klinpsy_KP$ABH, klinpsy_KP$CTd, klinpsy_KP$UTG, klinpsy_KP$ITd, sep = ". "), klinpsy_KP$text)
table(is.na(klinpsy_KP$text)) # n = 108649, 27640 missing

# 2nd case: abstracts and titles are in English -> use translations
table(klinpsy_KP$ABLN == "german" & is.na(klinpsy_KP$text)) # n = 1759
klinpsy_KP$text <- ifelse(klinpsy_KP$ABLN == "german" & is.na(klinpsy_KP$text), 
                        paste(klinpsy_KP$TIUE, klinpsy_KP$ABN, klinpsy_KP$CTd, klinpsy_KP$UTG, klinpsy_KP$ITd, sep = ". "), klinpsy_KP$text)
table(is.na(klinpsy_KP$text)) # n = 110408, 25881 missing

# 3rd case: no abstracts, document language is German
table(klinpsy_KP$LA == "german" & is.na(klinpsy_KP$text)) # n = 3254
klinpsy_KP$text <- ifelse(klinpsy_KP$LA == "german" & is.na(klinpsy_KP$text), 
                        paste(klinpsy_KP$`TI+TIU`, klinpsy_KP$CTd, klinpsy_KP$UTG, klinpsy_KP$ITd, sep = ". "), klinpsy_KP$text)
table(is.na(klinpsy_KP$text)) # n = 113662, 22626 missing

# 4th case: no abstracts, document language is English
table(klinpsy_KP$LA == "english" & is.na(klinpsy_KP$text)) # n = 22627
klinpsy_KP$text <- ifelse(klinpsy_KP$LA == "english" & is.na(klinpsy_KP$text), 
                        paste(klinpsy_KP$TIUE, klinpsy_KP$CTd, klinpsy_KP$UTG, klinpsy_KP$ITd, sep = ". "), klinpsy_KP$text)
table(is.na(klinpsy_KP$text)) # n = 136289, 1 missing

# check missings
klinpsy_KP$text <- ifelse(is.na(klinpsy_KP$text), 
                        paste(klinpsy_KP$`TI+TIU`, klinpsy_KP$CTd, klinpsy_KP$UTG, klinpsy_KP$ITd, sep = ". "), klinpsy_KP$text)
table(is.na(klinpsy_KP$text)) # n = 136289




### stopword list for German scientific abstracts by abi
stopwords <- read.csv("U:/R Projekte/Stopwords/stopwords-abi1019.csv", header = FALSE)
stopwords <- as.character(stopwords$V1)
# additional stopwords for this corpus
stopwords_klin <- c("translated", "by", "deepl", "abstract", "not", "released", "publisher", "relased",
                    "vortrag", "follow-up", "übersetzung")





#### corpus creation ----

# corpus
corpus <- corpus(klinpsy_KP$text)

# document-feature matrix
DFM <- corpus %>% 
  tokens() %>% 
  tokens_tolower() %>% 
  tokens_remove(pattern = c(stopwords, stopwords_klin, stopwords("en"), stopwords("de"))) %>%
  tokens(remove_punct = TRUE, remove_numbers = TRUE, remove_symbols = TRUE) %>% dfm()

topfeatures(DFM, n = 25)  

# Maier et al. (2018): In our empirical study, relative pruning was applied, 
# removing all terms that occurred in more than 99% or less than .5% of all documents 
# (Denny & Spirling, 2017; Grimmer, 2010; Grimmer & Stewart, 2013)
# inspect change in vocabulary size and dropped docs depending on min_docfreq


DFM <- dfm_trim(DFM, min_docfreq = 30, max_docfreq = round(0.99*dim(DFM)[1]))

# convert to package topicmodels
DTM <- convert(DFM, to = "topicmodels")
save(DTM, file = "DTM_94.RData")



#### model selection ----

source("U:/R Projekte/Forschungsinteressen/topic_coherence.R")
source("U:/R Projekte/Forschungsinteressen/reliability.R")

# grid search

# Following Maier et al. (2018), a variety of models with different parameters k and alpha is computed
# for each k, model with alpha highest topic coherence and exclusivity is selected
# fixed delta = 1/k

# The range of k and alpha determined by prior examinations (Bittermann & Fischer, 2018; Bittermann, 2019)
# experience shows that the higher k, the higher alpha needs to be in order to get high thetas for relevant topics
# in a doc (at least when using CT). For example, if a doc adresses only one topic, its theta should be near 100 % and all other doc-topic-probs
# near %. For k = 100, values of alpha should be smaller than 0.01. For k = 50, 0.01 would be fine.
# (the smaller alpha, the less topics per doc)

klist <- seq(30, 100, by = 5)
alphalist <- c(0.001, 0.01)
seedlist <- c(12, 4321, 28, 3323) #  4  random seeds (1 for each core in parallel computing)
cores <- 4 # parallel computing

lda_besta <- list() # saves the best models across all seeds. Best is defined using the mean of Semantic Coherence and Exclusivity.
lda_bestk_a <- list() # saves all lda_besta for each value of k
CO <- list()
EX <- list()
COEX_list <- list()
COEX_max <- list()


start <- Sys.time()
for (o in 1:length(klist)){
  k <- klist[o]
  
  for (i in 1:length(alphalist)){
    a <- alphalist[i]
    cluster <- parallel::makeCluster(cores)
    parallel::setDefaultCluster(cluster)
    parallel::clusterExport(varlist = c("DTM", "k", "a"), envir = environment())
    lda_temp <- parallel::parLapply(X = seedlist, fun = function(x) {
      topicmodels::LDA(DTM, k = k, method = "Gibbs", control = list(alpha = a, delta = 1/k, nstart = 1, seed = x))
    })
    parallel::stopCluster(cluster)
    
    for (j in 1:length(seedlist)) {
      EX[[j]] <- mean(quantqual::exclusivityLDA(lda_temp[[j]], as.matrix(DTM), num.terms = 100))
      CO[[j]] <- mean(tmca_topic_coherence(DTM, posterior(lda_temp[[j]])$terms, N = 100))
    }
    
    COEX_list[[i]] <- rowMeans(cbind(scale(unlist(EX)), scale(unlist(CO)))) # mean coherence and exclusivity (scaled)
    lda_besta[[i]] <- lda_temp[[which.max(unlist(COEX_list[[i]]))]] # save the best model of seeds for each alpha
    COEX_max[[i]] <- max(unlist(COEX_list[[i]]))
  }
  
  lda_bestk_a[[o]] <- lda_besta[[which.max(unlist(COEX_max))]] # save the best model of alphas for each k
  
}
end <- Sys.time()
end-start

save(lda_bestk_a, file = "lda_bestk_a.RData")
#load("lda_bestk_a.RData")


all_alphas <- list()
for (i in 1:length(lda_bestk_a)){
  all_alphas[[i]] <- lda_bestk_a[[i]]@alpha
}
unlist(all_alphas)


#### metrics ----

## evaluation metrics
LL_best <- list()
EX_best <- list()
CO_best <- list()


for (i in 1:length(klist)){
  LL_best[[i]] <- lda_bestk_a[[i]]@loglikelihood
  EX_best[[i]] <- mean(exclusivityLDA(lda_bestk_a[[i]], as.matrix(DTM), num.terms = 100))
  CO_best[[i]] <- mean(tmca_topic_coherence(DTM, posterior(lda_bestk_a[[i]])$terms, N = 100))
}

# correlations with alpha
cor(data.frame(unlist(all_alphas), unlist(LL_best), unlist(EX_best), unlist(CO_best)))

plot(scale(unlist(CO_best)), type = "l", col = "black", xaxt = "n", ylim = c(-1.5, 1.5), lwd = 2,
     main = "LDA Model Evaluation", ylab = paste("Scaled Score"), 
     xlab = "Number of Topics")
lines(scale(unlist(EX_best)), type = "l", col = "blue", lwd = 2)
lines(scale(unlist(LL_best)), type = "l", col = "purple", lwd = 2)
axis(1, at=1:length(klist), labels=klist)
legend("bottomright", c("Semantic Coherence", "Exclusivity", "Likelihood"), 
       col = c("black", "blue", "red"), lty = "solid", lwd = 2)

# indicates k = 45, 55, 70

k1 <- 45
k2 <- 55
k3 <- 70

lda_bestk_a[[which(klist == k1)]]@alpha
lda_bestk_a[[which(klist == k2)]]@alpha
lda_bestk_a[[which(klist == k3)]]@alpha



#### theta inspections ----

# check assigned topics for some docs

doc <- 4653 # randomly pick several docs
#par(mfrow=c(2,4))
#for (i in 1:length(klist)){
par(mfrow=c(1,3))
for (i in c(which(klist == k1), which(klist == k2), which(klist == k3))){
  plot(posterior(lda_bestk_a[[i]])$topics[doc,], type = "h", 
       main = paste("Doc", doc, "bei k =", klist[i], 
                    "Max_theta =", round(max(posterior(lda_bestk_a[[i]])$topics[doc,]), 4)), 
       ylab = "Theta", xlab = "Topic")
}
#klinpsy_KP[doc, 31] # doc content

k1_topics <- which(posterior(lda_bestk_a[[which(klist == k1)]])$topics[doc,] > 0.05) # get topic numbers
k2_topics <- which(posterior(lda_bestk_a[[which(klist == k2)]])$topics[doc,] > 0.05)
k3_topics <- which(posterior(lda_bestk_a[[which(klist == k3)]])$topics[doc,] > 0.05)

klinpsy_KP[doc, 31]; as.matrix(terms(lda_bestk_a[[which(klist == k1)]], 15))[,k1_topics] # topic contents
klinpsy_KP[doc, 31]; as.matrix(terms(lda_bestk_a[[which(klist == k2)]], 15))[,k2_topics] # topic contents
klinpsy_KP[doc, 31]; as.matrix(terms(lda_bestk_a[[which(klist == k3)]], 15))[,k3_topics] # topic contents

# evaluation of models:
# doc   k with good fit
# 1444    55, 70
# 6111    55, 70
# 1505    45, 55, 70
# 5505    55, 70
# 9808    45, 55, 70
# 487     70
# 10600   55
# 9595    70
# 128     45, 55
# 4653    45, 55, 70

rm(lda_best_k, lda_besta)




#### final model ----

k <- 70 # set k

# get alpha of best k
alpha <- lda_bestk_a[[which(klist == k)]]@alpha

# different seeds
seedlist_rel <- c(12, 4321, 28, 3323, 37323, 2721, 127722, 97039, 173, 317231, 98778) # 11 seeds (10 for reliability checks)

CO <- list()
EX <- list()

cores <- 6
cluster <- parallel::makeCluster(cores)
parallel::setDefaultCluster(cluster)
parallel::clusterExport(varlist = c("DTM", "k", "alpha"), envir = environment())
lda_temp <- parallel::parLapply(X = seedlist_rel, fun = function(x) {
  topicmodels::LDA(DTM, k = k, method = "Gibbs", control = list(alpha = alpha, delta=1/k, nstart=1, seed=x))
})
parallel::stopCluster(cluster)

for (j in 1:length(seedlist_rel)) {
  EX[[j]] <- mean(quantqual::exclusivityLDA(lda_temp[[j]], as.matrix(DTM), num.terms = 100))
  CO[[j]] <- mean(tmca_topic_coherence(DTM, posterior(lda_temp[[j]])$terms, N = 100))
}

COEX <- rowMeans(cbind(scale(unlist(EX)), scale(unlist(CO)))) # mean coherence and exclusivity (scaled)
lda_kp <- lda_temp[[which.max(COEX)]] # save the best model

save(lda_temp, file = "lda_temp.RData")
save(lda_kp, file = "lda_kp.RData")




#### topic labels ----

lda.theta <- posterior(lda_kp)$topics

# top 20 (a abit more than usual for post-stemming)
top10termsPerTopicProb <- terms(lda_kp, 20)
topicNames <- apply(top10termsPerTopicProb, 2, paste, collapse = ", ")
# change symbols back
topicNames <- gsub("_"," ", topicNames) # remove "_"
topicNames <- gsub("µ1","(", topicNames)
topicNames <- gsub("µ2",")", topicNames)
topicNames <- gsub("µ3","/", topicNames)
topicNames <- gsub("µ4","'", topicNames)


# data frame with topics, theta, and labels
topic_all <- data.frame(
  TopicNumber = 1:k,
  Prevalence = colMeans(lda.theta))
topic_all$terms <- topicNames

topic <- topic_all

#### topic reliability ----

# Repeat inference with different seeds.
# These models will be compared pairwise in order to determine model reliability.
# Which topics occur across different inference runs (with different seeds)?

source("U:/R Projekte/Forschungsinteressen/reliability.R") # by A. Niekler & G. Wiedemann
# Niekler, A.: Automatisierte Verfahren für die Themenanalyse nachrichtenorientierter Textquellen: 
# Dissertation zur Erlangung des akademischen Grades Doktor-Ingenieur (Dr.-Ing.) im Fachgebiet Informatik, 
# http://asv.informatik.uni-leipzig.de/publication/file/350/Niekler_Diss.pdf, (2016).

lda_temp <- lda_temp[-which.max(COEX)] # run only once!!!


topic_rel <- topic_all # df with topic numbers and terms

for (i in 1:length(lda_temp)){
  tm_fit1 <- fromTopicModelsPackage(lda_kp)
  tm_fit2 <- fromTopicModelsPackage(lda_temp[[i]])
  
  relCosine <- reliability_cosine(tm_fit1, tm_fit2, topWordsToMatch = 50, threshold = 0.5)
  #print(relCosine$reliability)
  topicmatches <- relCosine$ids
  #sort(topicmatches[,1]) # topic numbers (in LDA_final) that occur across inference runs
  
  rel_df <- data.frame(topicmatches[topicmatches[,1] > 0, 1], rep(1, length(topicmatches[topicmatches[,1] > 0, 1])))
  names(rel_df) <- c("TopicNumber", paste0("rel_",i))
  topic_rel <- dplyr::left_join(topic_rel, rel_df, by = "TopicNumber")
}

topic_rel$rel_sum <- rowSums(topic_rel[,4:length(topic_rel)], na.rm = TRUE)
table(topic_rel$rel_sum)

keep <- topic_rel[topic_rel$rel_sum >= 8, 1] # numbers of topics that occured in at least 80 % of other inference runs
length(keep)


# check representative docs
t <- 2 # for topic 2
temp <- names(head(sort(lda.theta[, t], decreasing = TRUE), 10))
topdocs <- as.numeric(gsub("text", "", temp))
klinpsy_KP[topdocs, 5] # column 5 is titles



# drop meaningless and uninterpretable topics
topic <- topic_all[keep,]
rownames(topic) <- NULL
View(topic)
drop <- c(24)
keep <- keep[-drop]



### keep only reliable topics ----

k <- length(keep)

lda.theta <- posterior(lda_kp)$topics[, keep]
lda.topterms <- as.matrix(terms(lda_kp, 20))[, keep]

dim(lda.theta)
dim(lda.topterms)
lda.theta[1:5,1:5]
lda.topterms[,1:5]


# new IDs

topic <- topic_all[keep,]
topic[,1] <- 1:k
topic$oldID <- rownames(topic)
colnames(lda.theta) <- paste(1:k)

# check representative docs
t <- 53
temp <- names(head(sort(lda.theta[, t], decreasing = TRUE), 20))
topdocs <- as.numeric(gsub("text", "", temp))
klinpsy_KP[topdocs, 5] # column 5 is titles



### Hot Topics ----

# original code by Ponweiser, M. (2012). Latent Dirichlet Allocation in R (Diploma Thesis). Vienna University of Economics and Business, Vienna.
# http://epub.wu.ac.at/3558/1/main.pdf


n.by.PY <- t(table(klinpsy_KP$PY))
rownames(n.by.PY) <- "Frequency"
plot(as.numeric(n.by.PY), type = "b", main = "Anzahl Publikationen je Jahr", xaxt = "n", ylab = "", xlab = "")
axis(1, at = 1:39, labels = 1980:2018)

#write.csv2(t(n.by.PY), file = "years_export.csv", row.names = FALSE)



years <- levels(factor(klinpsy_KP$PY)) # list of years

### time.series

theta_mean_by_year_by <- by(lda.theta, klinpsy_KP$PY, colMeans)
theta_mean_by_year <- do.call("rbind",theta_mean_by_year_by)
#colnames(theta_mean_by_year) = paste(keep)  # column names: only numbers (needed for plot)
theta_mean_by_year_ts <- ts(theta_mean_by_year, start = as.integer(years[1])) #Time-series
theta_mean_by_year_time <- time(theta_mean_by_year)

### slopes

#Linear model
theta_mean_lm <- apply(theta_mean_by_year, 2, function(x) lm(x ~ theta_mean_by_year_time)) # 2 is margin for columns
theta_mean_lm_coef <- lapply(theta_mean_lm,function(x) coef(summary(x)))
theta_mean_lm_coef_sign <- sapply(theta_mean_lm_coef,'[',"theta_mean_by_year_time","Pr(>|t|)")
theta_mean_lm_coef_slope <- sapply(theta_mean_lm_coef,'[',"theta_mean_by_year_time","Estimate")

# devide in positive and negative slopes
theta_mean_lm_coef_slope_pos <- theta_mean_lm_coef_slope[theta_mean_lm_coef_slope >= 0]
theta_mean_lm_coef_slope_neg <- theta_mean_lm_coef_slope[theta_mean_lm_coef_slope < 0]

# table of topics with positive and negative trends by level of significance
p_level <- c(0.05, 0.01, 0.001, 0.0001)
significance_total <- sapply(p_level, 
                             function(x) (theta_mean_lm_coef_sign[theta_mean_lm_coef_sign < x]))
significance_neg <- sapply(1:length(p_level),
                           function(x) intersect(names(theta_mean_lm_coef_slope_neg),names(significance_total[[x]])))
significance_pos <- sapply(1:length(p_level),
                           function(x) intersect(names(theta_mean_lm_coef_slope_pos),names(significance_total[[x]])))

### frequencies
outputmat <- rbind(sapply(significance_neg,length),
                   sapply(significance_pos,length ),
                   sapply(significance_total,length ))
rownames(outputmat) <- c("Negative trend", "Positive trend", "Total")
colnames(outputmat) <- c("p<0.05", "p<0.01", "p<0.001", "p<0.0001")
outputmat

### hot & cold

topics_hot <- as.numeric(names(sort(theta_mean_lm_coef_slope[significance_pos[[3]]], decreasing=TRUE)))
topics_cold <- as.numeric(names(sort(theta_mean_lm_coef_slope[significance_neg[[3]]], decreasing=FALSE)))

hot <- head(topics_hot, 10)

hot_and_cold_ts <- cbind(theta_mean_by_year_ts[,topics_hot[1:(length(topics_hot))]],
                         theta_mean_by_year_ts[,topics_cold[1:(length(topics_cold))]],
                         deparse.level=0)
colnames(hot_and_cold_ts) <- as.character(c(topics_hot[1:(length(topics_hot))],topics_cold[1:(length(topics_hot))]))


hot_ts <- theta_mean_by_year_ts[,topics_hot[1:(length(topics_hot))]]
cold_ts <- theta_mean_by_year_ts[,topics_cold[1:(length(topics_cold))]]
#colnames(hot_ts) <- as.character(topics_hot[1:10])
#colnames(cold_ts) <- as.character(topics_cold[1:10])


library(lattice)
xyplot(hot_ts*100,
       layout = c(5,3),
       col = "black",
       ylim = c(0,5.5),
       ylab = "Prävalenz (in Prozent)",
       xlab = "Jahr",
       type = c("l", "g", "r"),
       scales = list(x = list(alternating = FALSE)),
       panel = function(...) {
         panel.xyplot(...)
         panel.abline(100/70, lty = "dotted")},
       main = "Hot Topics"
)


write.csv2(topic, file = "topic.csv", row.names = FALSE)
write.csv2(topic[topics_hot,], file = "hot_topics.csv", row.names = FALSE)




print(xyplot(theta_mean_by_year_ts[,names(sort(theta_mean_lm_coef_slope, decreasing = TRUE))]*100, # sort by slope
             layout = c(4,5),
             col = "black",
             par.settings = list(strip.background=list(col="lightgrey")),
             alpha = 1 , # Strich-Transparenz
             ylim = c(0,15) ,
             # ylab = "Mittleres Theta",
             ylab = "Prävalenz (in Prozent)",
             xlab = "Jahr",
             type = c("l", "g"),
             # aspect = "xy",
             #auto.key = list(space = "right") , # Legende
             # auto.key = FALSE,
             scales = list(x = list(alternating = FALSE)),
             # par.settings = standard.theme(color = FALSE )
             main = "Zeitlicher Verlauf der Themenprävalenz"
))









